#include "dmlite_internal.h"

/**
 * Functions/macros internal to the plugin.
 */
#define LOGMSG_SIZE 1024

void dmlite_gfs_log(dmlite_handle_t * dmlite_handle_obj, globus_gfs_log_type_t log_type, char * format, ...) {
	va_list args;
	char buf[LOGMSG_SIZE];

	va_start(args, format);
	vsnprintf(buf, LOGMSG_SIZE, format, args);
	va_end(args);
	if (dmlite_handle_obj && dmlite_handle_obj->subject && dmlite_handle_obj->client_host) {
		globus_gfs_log_message(log_type, "dmlite :: %s :: %s :: %s\n",
			buf, dmlite_handle_obj->subject, dmlite_handle_obj->client_host);
	} else if (dmlite_handle_obj && dmlite_handle_obj->subject) {
		globus_gfs_log_message(log_type, "dmlite :: %s :: %s\n",
			buf, dmlite_handle_obj->subject);
	} else {
		globus_gfs_log_message(log_type, "dmlite :: %s\n", buf);
	}
}

void dmlite_gfs_hostid2host(char * host_id, char * host) {
	char * p;
	char * tmp = strdup(host_id);

	p = strrchr(tmp, ':');
	if(p) *p = '\0';

	host = strncpy(host, tmp, HOST_NAME_MAX);
	free(tmp);
}

const char * dmlite_gfs_fixpath(const char * path, globus_bool_t hostname) {
	char *p, *p2;

	while (*path == '/' && *(path + 1) == '/')
		path++;  /* reduce leading multiple slashes to one */

	/* first colon-slash */
	p = strstr(path, ":/");

	/* first slash after any leading slashes */
	p2 = strchr((*path == '/') ? (path + 1) : path, '/');

	/* treat it as a hostname:/path if there are no slashes before the colon-slash */
	if (p2 == p + 1) {
		if (hostname == GLOBUS_FALSE)
			path = p2;
		else if (*path == '/')
			path++;
	}

	return path;
}

globus_result_t dmlite_error2gfs_result(const char * name, dmlite_handle_t * handle, struct dmlite_context * context) {
	globus_result_t result = GLOBUS_SUCCESS;
	int dmlite_errnum;
	int posix_errno = -1;

	GlobusGFSName(name);

	dmlite_errnum = dmlite_errno(context);

	posix_errno = DMLITE_ERRNO(dmlite_errnum);

	if (DMLITE_ETYPE(dmlite_errnum) == DMLITE_USER_ERROR) {
		dmlite_gfs_log(handle, GLOBUS_GFS_LOG_WARN, "user error :: %d(%d) :: %s",
			dmlite_errnum, posix_errno, dmlite_error(context));
		result = GlobusGFSErrorSystemError(_gfs_name, posix_errno);
	} else {
		dmlite_gfs_log(handle, GLOBUS_GFS_LOG_ERR, "internal error :: %d(%d) :: %s",
			dmlite_errnum, posix_errno, dmlite_error(context));
		result = GlobusGFSErrorGeneric("Internal server error");
	}

	return result;
}

globus_result_t posix_error2gfs_result(const char * name, dmlite_handle_t * handle, int errnum, char * format, ...) {
	va_list args;
	globus_result_t result;
	char buf[LOGMSG_SIZE];

	GlobusGFSName(name);

	va_start(args, format);
	vsnprintf(buf, LOGMSG_SIZE, format, args);
	dmlite_gfs_log(handle, GLOBUS_GFS_LOG_ERR, "internal error :: %d :: %s", errnum, buf);
	va_end(args);

	if (errnum != EFAULT) {
		result = GlobusGFSErrorSystemError(_gfs_name, errnum);
	} else {
		result = GlobusGFSErrorGeneric("Internal server error");
	}

	return result;
}

void dmlite_stat2gfs(char * name, struct stat * stat, globus_gfs_stat_t * gfs_stat) {

	gfs_stat->name = strdup(name);
	gfs_stat->symlink_target = NULL;
	gfs_stat->dev = 0; // This field is not initialized by DMLite
	gfs_stat->ino = stat->st_ino;
	gfs_stat->mode = stat->st_mode;
	gfs_stat->nlink = stat->st_nlink;
	gfs_stat->uid = stat->st_uid;
	gfs_stat->gid = stat->st_gid;
	gfs_stat->size = stat->st_size;
	gfs_stat->atime = stat->st_atime;
	gfs_stat->ctime = stat->st_ctime;
	gfs_stat->mtime = stat->st_mtime;
}

struct dmlite_context * dmlite_get_context(dmlite_handle_t * dmlite_handle) {

	int retval = 0;
	struct dmlite_context * dmlite_context_obj = NULL;
	struct dmlite_credentials dmlite_creds_obj;

	/* Get a dmlite_manager object first */
	if (!dmlite_handle->manager) {
		dmlite_gfs_log(dmlite_handle, GLOBUS_GFS_LOG_ERR, "manager was null... this should not happen");
		goto errout;
	}

	/* Now we can fetch a new context */
	dmlite_context_obj = dmlite_context_new(dmlite_handle->manager);
	if (!dmlite_context_obj) {
		dmlite_gfs_log(dmlite_handle, GLOBUS_GFS_LOG_ERR, "failed to create context :: %s",
			dmlite_manager_error(dmlite_handle->manager));
		goto errout;
	}

	/* Set user credentials (only support GSI for the moment) */
	memset(&dmlite_creds_obj, 0, sizeof(dmlite_creds_obj));
	dmlite_creds_obj.mech = "ID"; //TODO: Use GSI here (configurable??)
	dmlite_creds_obj.client_name = dmlite_handle->subject;
	dmlite_creds_obj.remote_address = dmlite_handle->client_host;
	retval = dmlite_setcredentials(dmlite_context_obj, &dmlite_creds_obj);
	if (retval) {
		dmlite_gfs_log(dmlite_handle, GLOBUS_GFS_LOG_ERR, "failed to set credentials :: %s",
			dmlite_error(dmlite_context_obj));
		goto errout;
	}

	return dmlite_context_obj;

errout:
	if (dmlite_context_obj)
		dmlite_context_free(dmlite_context_obj);
	return NULL;
}

dmlite_fd * dmlite_gfs_open(struct dmlite_context * dmlite_context_obj, dmlite_handle_t * dmlite_handle_obj,
	const char * path, int flags) {

	dmlite_location *	dmlite_location_obj = NULL;
	dmlite_any_dict *	dmlite_query = NULL;
	const char *		rfn = dmlite_gfs_fixpath(path, GLOBUS_TRUE);
	const char *		localpath = dmlite_gfs_fixpath(path, GLOBUS_FALSE);
	int			mode;

	/* Sanity check */
	if (!path || !dmlite_context_obj || !dmlite_handle_obj)
		return NULL;

	/* At first we try to fetch a replica */
	if((flags & O_ACCMODE) == O_RDONLY) {
		mode = R_OK;
		dmlite_location_obj = dmlite_get(dmlite_context_obj, localpath);
	} else {
		mode = W_OK;
		dmlite_location_obj = dmlite_put(dmlite_context_obj, localpath);
	}

	if (dmlite_location_obj) {
		snprintf(dmlite_handle_obj->pfn, PATH_MAX, "%s:%s", dmlite_location_obj->chunks[0].url.domain, dmlite_location_obj->chunks[0].url.path);
		dmlite_handle_obj->is_replica = GLOBUS_TRUE;
		dmlite_query = dmlite_location_obj->chunks[0].url.query;
	} else if (dmlite_errno(dmlite_context_obj) == ENOENT) {
		dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "no replicas, trying RFN :: %s", rfn);
		if (dmlite_accessr(dmlite_context_obj, rfn, mode) && dmlite_errno(dmlite_context_obj) != DMLITE_NO_SUCH_REPLICA) {
			dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_ERR, "access to RFN denied :: %s :: %s",
					rfn, dmlite_error(dmlite_context_obj));
			goto errout;
		}
		dmlite_handle_obj->is_replica = GLOBUS_FALSE;
		strncpy(dmlite_handle_obj->pfn, rfn, PATH_MAX);
	} else {
		dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_ERR, "failed to fetch replica :: %s :: %s",
				localpath, dmlite_error(dmlite_context_obj));
		goto errout;
	}
	dmlite_handle_obj->location = dmlite_location_obj;
	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "opening :: %s", dmlite_handle_obj->pfn);
	return dmlite_fopen(dmlite_context_obj, dmlite_handle_obj->pfn, flags | O_INSECURE | O_LARGEFILE, dmlite_query);

errout:
	if (dmlite_location_obj)
		dmlite_location_free(dmlite_location_obj);
	return NULL;
}

int dmlite_gfs_close(struct dmlite_context * dmlite_context_obj, dmlite_handle_t * dmlite_handle_obj, globus_bool_t ok) {

	int retval = 0;

	/* Sanity check 1 */
	if (!dmlite_handle_obj || !dmlite_handle_obj->fd)
		return -1;

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "closing :: %s", dmlite_handle_obj->pfn);

	retval = dmlite_fclose(dmlite_handle_obj->fd);

	if (!dmlite_handle_obj->is_replica)
		return retval;

	if (!dmlite_context_obj || !dmlite_handle_obj->location)
		return -1;

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "replica put status :: %d", ok);

	if (ok)
		retval = dmlite_donewriting(dmlite_context_obj, dmlite_handle_obj->location);
	else
		retval = dmlite_put_abort(dmlite_context_obj, dmlite_handle_obj->location);

	return retval;
}
